//===--- clangserver.cpp - C-Language Front-end compile server-------------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
//  This utility may be invoked in the following manner:
//   clangserver --help                - Output help info.
//
//===----------------------------------------------------------------------===//

#include "clang.h"
#include "ASTConsumers.h"
#include "DiagnosticBuffer.h"
#include "HTMLDiagnostics.h"
#include "NetSession.h"
#include "clang/Analysis/PathDiagnostic.h"
#include "clang/AST/TranslationUnit.h"
#include "clang/CodeGen/ModuleBuilder.h"
#include "clang/Sema/ParseAST.h"
#include "clang/AST/ASTConsumer.h"
#include "clang/Parse/Parser.h"
#include "clang/Lex/HeaderSearch.h"
#include "clang/Basic/FileManager.h"
#include "clang/Basic/SourceManager.h"
#include "clang/Basic/TargetInfo.h"
#include "llvm/Module.h"
#include "llvm/ADT/SmallPtrSet.h"
#include "llvm/Bitcode/Serialize.h"
#include "llvm/Bitcode/Deserialize.h"
#include "llvm/Bitcode/ReaderWriter.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Support/MemoryBuffer.h"
#include "llvm/System/Signals.h"
#include "llvm/Config/config.h"
#include "llvm/ADT/OwningPtr.h"
#include "llvm/System/Path.h"
#include <signal.h>
#include <vector>
#include <memory>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <pthread.h> 
using namespace clang;

struct ThreadData {
  pthread_t PThread;
  int Connection; // This is the thread ID
  struct sockaddr_in Address;
};

static std::vector<ThreadData*> ThreadsData;
static pthread_mutex_t ThreadsData_lock = PTHREAD_MUTEX_INITIALIZER;

//===----------------------------------------------------------------------===//
// Global options.
//===----------------------------------------------------------------------===//

static llvm::cl::opt<bool>
Verbose("v", llvm::cl::desc("Enable verbose output"));
static llvm::cl::opt<bool>
Stats("print-stats", 
      llvm::cl::desc("Print performance metrics and statistics"));

//===----------------------------------------------------------------------===//
// Server options.
//===----------------------------------------------------------------------===//

static llvm::cl::opt<unsigned> Port("port",
       llvm::cl::desc("Port of remote compiler service"),
       llvm::cl::init(50000));
	   
static llvm::cl::opt<unsigned>
QueueSize("queuesize",
      llvm::cl::desc("Server request queue size"),
      llvm::cl::init(5));
	  
static llvm::cl::opt<unsigned>
MaxClient("maxclient",
      llvm::cl::desc("Maximum number of connected clients"),
      llvm::cl::init(5));
	  
//===----------------------------------------------------------------------===//
// Main driver
//===----------------------------------------------------------------------===//

/// CreateASTConsumer - Create the ASTConsumer for the corresponding program
///  action.  These consumers can operate on both ASTs that are freshly
///  parsed from source files as well as those deserialized from Bitcode.
static ASTConsumer* CreateASTConsumer(const std::string& InFile,
                                      Diagnostic& Diag, FileManager& FileMgr, 
                                      const LangOptions& LangOpts,
                                      Preprocessor *PP,
                                      llvm::Module *&DestModule,
                                      ProgActions ProgAction,
                                      std::ostream *StdOut) {

  switch (ProgAction) {
  default:
    return NULL;
    break;
      
  case ParseSyntaxOnly:              // -fsyntax-only, no output
    return new ASTConsumer();
      
  case ASTPrint:
    return CreateASTPrinter(StdOut);
      
  case EmitHTML:
    //return CreateHTMLPrinter(StdOut, Diag, PP, PPF);
    break;

  case EmitLLVM:
  case EmitBC:
    DestModule = new llvm::Module(InFile);
    //return CreateLLVMCodeGen(Diag, LangOpts, DestModule, GenerateDebugInfo);
    return CreateLLVMCodeGen(Diag, LangOpts, DestModule, false);
    break;

  case SerializeAST:
    // FIXME: Allow user to tailor where the file is written.
    return CreateASTSerializer(InFile, *StdOut, Diag);
    break;
      
  case RewriteObjC:
    return CreateCodeRewriterTest(InFile, *StdOut, Diag, LangOpts);
    break;
      
  case RunAnalysis:
    // Emits more than one file.
    //assert (!AnalysisList.empty());
    //Consumer = CreateAnalysisConsumer(&AnalysisList[0],
    //                                  &AnalysisList[0]+AnalysisList.size(),
    //                                  Diag, PP, PPF, LangOpts,
    //                                  AnalyzeSpecificFunction,
    //                                  OutputFile, VisualizeEG, TrimGraph,
    //                                  AnalyzeAll);
    break;
  }
  return NULL;
}

/// ProcessInputFile - Process a single input file with the specified state.
///
static void ProcessInputFile(Preprocessor &PP, const std::string &InFile,
                             ProgActions ProgAction, std::ostream *StdOut) {
  ASTConsumer* Consumer = NULL;
  llvm::Module *CodeGenModule = 0;
  
  Consumer = CreateASTConsumer(InFile, PP.getDiagnostics(),
                               PP.getFileManager(), PP.getLangOptions(),
                               &PP, CodeGenModule,
							   ProgAction, StdOut);
  if (!Consumer) {      
    fprintf(stderr, "Unsupported program action!\n");
	exit(1);
  }
  // This deletes Consumer.
  ParseAST(PP, Consumer, Stats);

  // If running the code generator, finish up now.
  if (CodeGenModule) {
    if (ProgAction == EmitLLVM) {
      CodeGenModule->print(StdOut);
    } else {
      assert(ProgAction == EmitBC);
      llvm::WriteBitcodeToFile(CodeGenModule, *StdOut);
    }
    delete CodeGenModule;
  }
  /*
  if (Stats) {
    fprintf(stderr, "\nSTATISTICS FOR '%s':\n", InFile.c_str());
    PP.PrintStats();
    PP.getIdentifierTable().PrintStats();
    PP.getHeaderSearchInfo().PrintStats();
    PP.getSourceManager().PrintStats();
    fprintf(stderr, "\n");
  }
  */
}

static void ProcessSerializedFile(llvm::MemoryBuffer& BufferIn,
                                  Diagnostic& Diag,
                                  FileManager& FileMgr,
                                  ProgActions ProgAction,
                                  std::ostream *StdOut) {

  llvm::OwningPtr<TranslationUnit> TU(ReadASTBitcodeBuffer(BufferIn, FileMgr));
  std::string InFile = BufferIn.getBufferIdentifier();

  if (!TU) {
    //ErrorMsg = "error: file '"+InFile+"' could not be deserialized";
    return;
  }
  
  // Observe that we use the source file name stored in the deserialized
  // translation unit, rather than InFile.
  llvm::Module *CodeGenModule;

  ASTConsumer *Consumer = CreateASTConsumer(InFile, Diag, FileMgr,
                                            TU->getLangOptions(), 0,
                                            CodeGenModule, ProgAction, StdOut);
  
  if (!Consumer) {      
    fprintf(stderr, "Unsupported program action with serialized ASTs!\n");
	exit(1);
  }
  
  Consumer->Initialize(TU->getContext());
  
  // FIXME: We need to inform Consumer about completed TagDecls as well.
  for (TranslationUnit::iterator I=TU->begin(), E=TU->end(); I!=E; ++I)
    Consumer->HandleTopLevelDecl(*I);

  delete Consumer;

  if (CodeGenModule) {
    if (ProgAction == EmitLLVM) {
      CodeGenModule->print(StdOut);
    } else {
      assert(ProgAction == EmitBC);
      llvm::WriteBitcodeToFile(CodeGenModule, *StdOut);
    }
    delete CodeGenModule;
  }
}

static llvm::MemoryBuffer* GetRemoteFileBuffer(llvm::Deserializer &PacketIn,
                                               unsigned char *FilesData) {
  // File header: RemoteAction, FileName, FileSize,
  //              RemoteFileType, FileOffset
  RemoteActions RemoteActions = (clang::RemoteActions)PacketIn.ReadInt();
  std::string FName = PacketIn.ReadCStr();
  int FSize = PacketIn.ReadInt();
  RemoteFileType FType = (clang::RemoteFileType)PacketIn.ReadInt();
  int FOffset = PacketIn.ReadInt();

  return llvm::MemoryBuffer::getMemBuffer((const char*)FilesData+FOffset,
                                          (const char*)FilesData+FOffset+FSize,
                                          FName.c_str());
}

static void EmitDiagnostics(DiagnosticBuffer &DiagClient,
                            llvm::Serializer &PacketOut) {
  PacketOut.EmitInt(RemoteDiagnostic);
  // Diagnostic Num
  PacketOut.EmitInt(DiagClient.DiagItems.size());
  
  for (std::vector<DiagnosticBuffer::DiagItem>::iterator
       i = DiagClient.DiagItems.begin() ;
       i != DiagClient.DiagItems.end() ; ++i) {

    // DiagID
    PacketOut.EmitInt((*i).DiagID);
/*
    // Pos
    PacketOut.Emit((*i).Pos.getLocation());
    // Strs
    PacketOut.EmitInt((*i).Strs.size());
    for (std::vector<std::string>::iterator j = (*i).Strs.begin() ;
         j != (*i).Strs.end() ; ++j) {
      PacketOut.EmitCStr((*j).c_str());
    }
    // Ranges
    PacketOut.EmitInt((*i).Ranges.size());
    for (std::vector<SourceRange>::iterator j = (*i).Ranges.begin() ;
         j != (*i).Ranges.end() ; ++j) {
      PacketOut.Emit((*j));
    }
*/
  }
}

static void *ServeRequest(void *Data) {
  ThreadData *TData = (ThreadData*)Data;
  NetSession Session(TData->Connection, TData->Address);
  
  llvm::MemoryBuffer *FileBuffer;
  std::ostringstream StdOut(std::ios_base::out|
                            std::ios_base::in|
                            std::ios_base::binary);
  std::string InFile;
  //std::string ErrorMsg;
  
  // DEBUG
  fprintf(stderr, " * Starting new thread\n");

  // Read incoming request packet.
  llvm::MemoryBuffer *BufferIn = Session.GetPacket();
  if (BufferIn == NULL) {
    if (Verbose)
      fprintf(stderr, "%s\n", Session.GetErrorString().c_str());
    if (!Session.Close() && Verbose)
      fprintf(stderr, "%s\n", Session.GetErrorString().c_str());
    TData->Connection = -1;
    return NULL;
  }
  
  unsigned char *BufPtr = (unsigned char *) BufferIn->getBufferStart();
  u_long HeaderSize = ntohl(*((u_long*)BufPtr));
  BufPtr += sizeof(u_long);
  unsigned char *FilesData = BufPtr + HeaderSize;
  llvm::BitstreamReader StreamIn(BufPtr, BufPtr+HeaderSize);

  llvm::Deserializer PacketIn(StreamIn);
  PacketIn.SkipToBlock(0);
  
  // Parse incoming request packet.
  RemoteActions RemoteAction = (clang::RemoteActions)PacketIn.ReadInt();
  ProgActions ProgAction = (clang::ProgActions)PacketIn.ReadInt();
  std::string OutputFile = PacketIn.ReadCStr();

  // DEBUG
  fprintf(stderr, " * RemoteAction: %d\n", RemoteAction);
  fprintf(stderr, " * ProgAction: %d\n", ProgAction);
  fprintf(stderr, " * OutputFile: %s\n", OutputFile.c_str());

  // Setup clang

  // Configure our handling of diagnostics.
  llvm::OwningPtr<DiagnosticBuffer> DiagClient(new DiagnosticBuffer());
  Diagnostic Diags(*DiagClient);

  // Create a file manager object to provide access to and cache the filesystem.
  FileManager FileMgr;

  if (RemoteAction == RemoteProcessSerializedFile) {
    // Get file buffer from packet.
    FileBuffer = GetRemoteFileBuffer(PacketIn,FilesData);
    InFile = FileBuffer->getBufferIdentifier();
    ProcessSerializedFile(*FileBuffer, Diags, FileMgr, ProgAction, &StdOut);
    delete FileBuffer;
  }
  else {
    // Get information about the target being compiled for.
    std::string Triple = PacketIn.ReadCStr();
    llvm::OwningPtr<TargetInfo> Target(TargetInfo::CreateTargetInfo(Triple));
  
    // Initialize language options from packet.
    LangOptions LangInfo;
    PacketIn.Read(LangInfo);

    // Get predefines from packet.
    std::string Predefines = PacketIn.ReadCStr();

    // Get MainFile from packet.
    FileBuffer = GetRemoteFileBuffer(PacketIn,FilesData);
    InFile = FileBuffer->getBufferIdentifier();

    SourceManager SourceMgr;
    HeaderSearch HeaderInfo(FileMgr);
    Preprocessor PP(Diags, LangInfo, *Target, SourceMgr, HeaderInfo);
    PP.setPredefines(Predefines);
    // SourceManager takes ownership of FileBuffer
    SourceMgr.createMainFileIDForMemBuffer(FileBuffer);

    ProcessInputFile(PP, InFile, ProgAction, &StdOut);

    if (Stats)
      SourceMgr.PrintStats();

  } 

  // Write out the output.
  StdOut.rdbuf()->pubseekoff(0,std::ios_base::beg,std::ios_base::in);
  StdOut.rdbuf()->pubseekoff(0,std::ios_base::end,std::ios_base::out);

  // DEBUG
  fprintf(stderr, " * OutputSize: %d\n", StdOut.rdbuf()->in_avail());
  fprintf(stderr, " * StdOut: %s\n", StdOut.str().c_str());

  unsigned NumDiagnostics = Diags.getNumDiagnostics();
  fprintf(stderr, " * Errors: %d\n", Diags.getNumErrors());
  fprintf(stderr, " * %d diagnostic%s generated.\n", NumDiagnostics,
         (NumDiagnostics == 1 ? "" : "s"));

  if (StdOut.rdbuf()->in_avail() > 0) {
    // Create StdOut output or OutputFile
    
	// Some dirty hack to modify output file name.
    if ((ProgAction == EmitLLVM || ProgAction == EmitBC) && OutputFile == "") {
      llvm::sys::Path Path(InFile);
      Path.eraseSuffix();
      if (ProgAction == EmitLLVM)
        Path.appendSuffix("ll");
      else if (ProgAction == EmitBC)
        Path.appendSuffix("bc");
      OutputFile = Path.toString();
    }
	
    if (OutputFile == "" || OutputFile == "-")
      Session.PacketAddFile(OutputFile.c_str(), &StdOut, RemoteStdOutT);
    else
      Session.PacketAddFile(OutputFile.c_str(), &StdOut, RemoteFileT);
  }
  
  std::vector<unsigned char> BufferOut;
  llvm::BitstreamWriter StreamOut(BufferOut);
  llvm::Serializer PacketOut(StreamOut);  
  PacketOut.EnterBlock(0);  

  // write diagnostic results to packet
  EmitDiagnostics(*DiagClient, PacketOut);
  
  // Close result packet with RemoteComplete
  PacketOut.EmitInt(RemoteComplete);  
  
  // send result packet
  Session.SetPacketHeader(&PacketOut, &BufferOut);
  if (!Session.SendPacket()) {
    if (Verbose)
      fprintf(stderr, "%s\n", Session.GetErrorString().c_str());
  }
  if (!Session.Close() && Verbose)
    fprintf(stderr, "%s\n", Session.GetErrorString().c_str());
  /*
  if (Stats) {
    FileMgr.PrintStats();
    fprintf(stderr, "\n");
  }
  */
  delete BufferIn;

  // This inicates that this thread is dead.
  TData->Connection = -1;
  return NULL;
}

static const int SOCKET_ERROR = -1;
static int ServerSocket;

void DoCleanup() {
  // Kill all running threads.
  int KilledNum = 0;
  ThreadData *TD = 0;
  
  // Cleanup was done.
  if(ServerSocket < 0)
    return;
	
  pthread_mutex_lock(&ThreadsData_lock);

  while (ThreadsData.size() > 0) {
    TD = *ThreadsData.begin();
    if (TD->Connection == -1) {
      delete TD;
    } else {
      pthread_kill(TD->PThread,SIGKILL);
      KilledNum++;

      // Close threads socket
      if (close(TD->Connection) == SOCKET_ERROR) {
        if (Verbose)
          fprintf(stderr, "Could not close socket.\n");
      }
      delete TD;
      ThreadsData.erase(ThreadsData.begin());
    }
  }
  pthread_mutex_unlock(&ThreadsData_lock);

  if (Verbose)
    fprintf(stderr, "%d running threads was killed.\n", KilledNum);
  if (close(ServerSocket) == SOCKET_ERROR) {
    if (Verbose)
      fprintf(stderr, "Could not close socket.\n");
  }
  ServerSocket = -2;
}

int main(int argc, char **argv) {
  char IDBuffer[100];
  int IDLen = strlen(NetSession::ServiceID());
  ThreadData *TD;
  char Accept;
  struct sockaddr_in Address;
  int AddressSize = sizeof(struct sockaddr_in);
  int reuse_addr = 1;

  llvm::cl::ParseCommandLineOptions(argc, argv, " llvm clang cfe server\n");
  llvm::sys::PrintStackTraceOnErrorSignal();
  
  // Setup server.
  ServerSocket = socket(AF_INET,SOCK_STREAM,0);

  if (ServerSocket == SOCKET_ERROR) {
    fprintf(stderr, "Could not make a socket.\n");
    exit(1);
  }
  setsockopt(ServerSocket, SOL_SOCKET, SO_REUSEADDR, &reuse_addr,
             sizeof(reuse_addr));
			 
  Address.sin_addr.s_addr = INADDR_ANY;
  Address.sin_port = htons(Port);
  Address.sin_family = AF_INET;

  if (bind(ServerSocket, (struct sockaddr*)&Address, sizeof(Address))
      == SOCKET_ERROR) {
    fprintf(stderr, "Could not connect to host.\n");
    exit(1);
  }
  
  getsockname(ServerSocket, (struct sockaddr*)&Address,
             (socklen_t*)&AddressSize);

  if (listen(ServerSocket, QueueSize) == SOCKET_ERROR) {
    fprintf(stderr, "Could not listen for clients.\n");
    exit(1);
  }
  
  atexit (DoCleanup);
  llvm::sys::SetInterruptFunction(DoCleanup);
  // Main loop.
  while (true) {
    int Connection = accept(ServerSocket, (struct sockaddr*)&Address,
                            (socklen_t *)&AddressSize);

    // Exchange service ID and version
    Accept = 1;
    if (Connection == SOCKET_ERROR)
      Accept = 0;

    if (Accept && NetSession::SendAll(Connection, (char*)NetSession::ServiceID(),
                               IDLen) != IDLen)
      Accept = 0;

    if (Accept && NetSession::RecvAll(Connection, (char*)&IDBuffer, IDLen) != IDLen)
      Accept = 0;

    if (Accept && strncmp(NetSession::ServiceID(),IDBuffer,IDLen) != 0)
      Accept = 0;
	  
    if (!Accept) {
      // CONTROL-C means normal exit.
      if (Connection < 0) {
        if (Verbose)
          fprintf(stderr, "Exit.\n");
        break;
      }
      if (Verbose)
        fprintf(stderr, "Invalid request.\n");
      if (close(Connection) == SOCKET_ERROR) {
        if (Verbose)
          fprintf(stderr, "Could not close client socket.\n");
        exit(1);
      }
      continue;
    }

    // Accept request if we have free slot.
    pthread_mutex_lock(&ThreadsData_lock);
    // Remove dead threads.
    for (std::vector<ThreadData*>::iterator i = ThreadsData.begin();
         i != ThreadsData.end() ; ) {
      if ((*i)->Connection == -1) {
        delete (*i);
        ThreadsData.erase(i);
        i = ThreadsData.begin();
      }
      else {
        i++;
      }
    }
    if (Accept && ThreadsData.size() < MaxClient) {
      TD = new ThreadData;
      TD->Connection = Connection;
      TD->Address = Address;
      ThreadsData.push_back(TD);
    }
    else {
      Accept = 0;
    }
    // DEBUG
    fprintf(stderr, "** Jobs Count: %d  MaxJobNum: %d\n",
            ThreadsData.size(), (int)MaxClient);
    pthread_mutex_unlock(&ThreadsData_lock);

    if (NetSession::SendAll(Connection, &Accept, 1) != 1) {
      if (Verbose)
        fprintf(stderr, "Communication error.\n");
      if (close(Connection) == SOCKET_ERROR) {
        if (Verbose)
          fprintf(stderr, "Could not close client socket.\n");
        exit(1);
      }
    }

    // Start compiler thread.
    if (Accept) {
      pthread_create(&TD->PThread, NULL, ServeRequest, (void*)TD);
    }
    else {
      if (close(Connection) == SOCKET_ERROR) {
        if (Verbose)
          fprintf(stderr, "Could not close client socket.\n");
        exit(1);
      }
    }
  }

  // Close server socket.
  DoCleanup();
  return 0;
}
